`timescale 1ns / 1ps

module regs
(
	input i_clk,
	input i_reset,
	
	output o_vblank_nmi,
	
	// CPU interface
	input [2:0] i_cpu_addr,
	input i_cpu_write,
	input i_cpu_read,
	input [7:0] i_cpu_write_data,
	output reg [7:0] o_cpu_read_data,
	output reg o_cpu_datavalid,
	
	// VRAM interface
	output reg [13:0] o_vram_addr,
	output reg o_vram_read,
	output reg o_vram_write,
	input [7:0] i_vram_read_data,
	output reg [7:0] o_vram_write_data,
	
	// To palette memory
	output [4:0] o_pal_addr,
	output [5:0] o_pal_wr_data,
	output o_pal_write,
	input [5:0] i_pal_rd_data,
	
	// To SPR-RAM
	output [7:0] o_spr_addr,
	output [7:0] o_spr_wr_data,
	output o_spr_write,
	input [7:0] i_spr_rd_data,
	
	// Renderer
	input i_hscroll_cnt_load,
	input i_all_cnt_load,
	input i_hscroll_cnt_enable,
	input i_vscroll_cnt_enable,
	input i_rend_nt_read,
	input i_rend_at_read,
	input i_rend_bpt_read,
	input i_rend_bpt_bit3,
	input i_rend_opt_read,
	input i_rend_opt_bit3,
	input i_rend_opt_ptab,
	input [2:0] i_rend_opt_fv,
	input [7:0] i_rend_opt_tilenum,
	output [1:0] o_attrib,
	output [2:0] o_fine_h,
	input i_rend_set_vblank,
	input i_rend_clear_vblank,
	input i_rend_spritelimit,
	input i_rend_primaryhit,
	output o_object_pat_table,
	output o_object_big,
	output o_object_clip,
	output o_object_enable,
	output o_pf_clip,
	output o_pf_enable
);


// 2000 status registers
reg inc_port7_by_32; // 2
reg object_pat_table; // 3
reg objects_16_scanlines; // 5
reg vblank_enable; // 7

// 2001 status registers
reg show_left_8pix_playfield; // 1
reg show_left_8pix_objects; // 2
reg enable_playfield; // 3
reg enable_objects; // 4

// 2002 status registers
reg vblank_in_progress;
reg primaryhit;
reg spritelimit;

// Used for accessing 2005/2006
reg cpu_secondwrite;

// Address registers
reg [2:0] FV;
reg V;
reg H;
reg [4:0] VT;
reg [4:0] HT;
reg [2:0] FH;
reg S;
reg [7:0] PAR;
reg [1:0] AR;

// Address counters, loaded from subset of address registers
reg [2:0] FV_counter;
reg V_counter;
reg H_counter;
reg [4:0] VT_counter;
reg [4:0] HT_counter;

// used for loading data into PAR and AR after VRAM read data arrives
reg [2:0] rend_at_shift_amt;
reg rend_nt_read_r;
reg rend_at_read_r;

// Data from VRAM->CPU gets pipelined one cycle (first read returns garbage)
reg [7:0] vram_to_cpu_data_reg;

// Address for accessing SPR-RAM by CPU (gets auto-incremented after 2004 access)
reg [7:0] cpu_spr_addr;


// Vblank interrupt generation
assign o_vblank_nmi = vblank_enable & vblank_in_progress;

// Is the CPU reading/writing the palette?
wire cpu_access_pal = o_vram_addr[13:8] == 6'h3F;

// Is the CPU reading or writing to 0x2007? + delayed version
wire cpu_access_2007 = (i_cpu_read | i_cpu_write) && i_cpu_addr == 3'd7;
reg cpu_access_2007_r;

// Delayed versions of CPU signals used for read capture
reg cpu_read_r;
reg [2:0] cpu_addr_r;

// Connect CPU to palette memory
assign o_pal_addr = o_vram_addr[4:0];
assign o_pal_wr_data = i_cpu_write_data[5:0];
assign o_pal_write = i_cpu_write && cpu_access_pal && i_cpu_addr == 3'd7;

// Connect CPU to SPR RAM
assign o_spr_addr = cpu_spr_addr;
assign o_spr_write = i_cpu_write && i_cpu_addr == 3'd4;
assign o_spr_wr_data = i_cpu_write_data;

// Give renderer the value of some registers
assign o_attrib = AR;
assign o_fine_h = FH;
assign o_object_pat_table = object_pat_table;
assign o_object_big = objects_16_scanlines;
assign o_object_clip = ~show_left_8pix_objects;
assign o_object_enable = enable_objects;
assign o_pf_clip = ~show_left_8pix_playfield;
assign o_pf_enable = enable_playfield;



// Synchronous control of address and status registers
always @ (posedge i_clk or posedge i_reset) begin
	if (i_reset) begin
		inc_port7_by_32 <= 'd0;
		object_pat_table <= 'd0;
		objects_16_scanlines <= 'd0;
		vblank_enable <= 'd0;
		show_left_8pix_playfield <= 'd1;
		show_left_8pix_objects <= 'd1;
		enable_playfield <= 'd1;
		enable_objects <= 'd1;
		cpu_secondwrite <= 'd0;
		FV <= 'd0;
		V <= 'd0;
		H <= 'd0;
		VT <= 'd0;
		HT <= 'd0;
		FH <= 'd0;
		S <= 'd1;
		cpu_spr_addr <= 'd0;
		cpu_addr_r <= 'd0;
		cpu_read_r <= 'd0;
	end
	else begin
		cpu_read_r <= i_cpu_read;
		cpu_addr_r <= i_cpu_addr;
		
		case (i_cpu_addr)
			3'd0 : if (i_cpu_write) begin
				H <= i_cpu_write_data[0];
				V <= i_cpu_write_data[1];
				inc_port7_by_32 <= i_cpu_write_data[2];
				object_pat_table <= i_cpu_write_data[3];
				S <= i_cpu_write_data[4];
				objects_16_scanlines <= i_cpu_write_data[5];
				vblank_enable <= i_cpu_write_data[7];
			end
			
			3'd1 : if (i_cpu_write) begin
				show_left_8pix_playfield <= i_cpu_write_data[1];
				show_left_8pix_objects <= i_cpu_write_data[2];
				enable_playfield <= i_cpu_write_data[3];
				enable_objects <= i_cpu_write_data[4];
			end
			
			3'd2 : if (i_cpu_read) begin
				cpu_secondwrite <= 1'b0;
				vblank_in_progress <= 1'b0;
			end
			
			3'd3 : if (i_cpu_write) begin
				cpu_spr_addr <= i_cpu_write_data;
			end
			
			3'd4 : if (i_cpu_read || i_cpu_write) begin
				cpu_spr_addr <= cpu_spr_addr + 8'd1;
			end
			
			3'd5 : if (i_cpu_write) begin
				if (cpu_secondwrite == 1'b0) begin
					FH <= i_cpu_write_data[2:0];
					HT <= i_cpu_write_data[7:3];
					cpu_secondwrite <= 1'b1;
				end
				else begin
					FV <= i_cpu_write_data[2:0];
					VT <= i_cpu_write_data[7:3];
					cpu_secondwrite <= 1'b0;
				end
			end
			
			3'd6 : if (i_cpu_write) begin
				if (cpu_secondwrite == 1'b0) begin
					VT[4:3] <= i_cpu_write_data[1:0];
					H <= i_cpu_write_data[2];
					V <= i_cpu_write_data[3];
					FV <= {1'b0, i_cpu_write_data[5:4]};
					cpu_secondwrite <= 1'b1;
				end
				else begin
					HT <= i_cpu_write_data[4:0];
					VT[2:0] <= i_cpu_write_data[7:5];
					cpu_secondwrite <= 1'b0;
				end
			end
		endcase

		if (!i_cpu_read || i_cpu_addr != 3'd2) begin
			if (i_rend_set_vblank) vblank_in_progress <= 1'b1;
			else if (i_rend_clear_vblank) vblank_in_progress <= 1'b0;
		end
		
		if (i_rend_spritelimit) spritelimit <= 1'b1;
		else if (i_rend_set_vblank) spritelimit <= 1'b0;
		
		if (i_rend_primaryhit) primaryhit <= 1'b1;
		else if (i_rend_set_vblank) primaryhit <= 1'b0;
	end
end


// Data capture from VRAM into registers
always @ (posedge i_clk or posedge i_reset) begin
	if (i_reset) begin
		cpu_access_2007_r <= 'd0;
		vram_to_cpu_data_reg <= 'd0;
		rend_at_shift_amt <= 'd0;
		rend_nt_read_r <= 'd0;
		rend_at_read_r <= 'd0;
		PAR <= 'd0;
		AR <= 'd0;
	end
	else begin
		cpu_access_2007_r <= cpu_access_2007;
		rend_nt_read_r <= i_rend_nt_read;
		rend_at_read_r <= i_rend_at_read;
		rend_at_shift_amt <= {VT_counter[1], HT_counter[1], 1'b0};
		
		if (cpu_access_2007_r) begin
			vram_to_cpu_data_reg <= i_vram_read_data;
		end
		
		if (rend_nt_read_r) begin
			PAR <= i_vram_read_data;
		end
		
		if (rend_at_read_r) begin
			AR <= i_vram_read_data >> rend_at_shift_amt;
		end
	end
end

// VRAM interface control
always @* begin
	o_vram_read = 'bx;
	o_vram_write = 'b0;
	o_vram_addr = 'bx;
	o_vram_write_data = 'bx;
	
	if (cpu_access_2007) begin
		o_vram_read = i_cpu_read & ~cpu_access_pal;
		o_vram_write = i_cpu_write & ~cpu_access_pal;
		o_vram_addr = {FV_counter[1:0], V_counter, H_counter, VT_counter, HT_counter};
		o_vram_write_data = i_cpu_write_data;
	end
	else if (i_rend_nt_read) begin
		o_vram_read = 1'b1;
		o_vram_addr = {2'b10, V_counter, H_counter, VT_counter, HT_counter};
	end
	else if (i_rend_at_read) begin
		o_vram_read = 1'b1;
		o_vram_addr = {2'b10, V_counter, H_counter, 4'b1111, VT_counter[4:2], HT_counter[4:2]};
	end
	else if (i_rend_bpt_read) begin
		o_vram_read = 1'b1;
		o_vram_addr = {1'b0, S, PAR, i_rend_bpt_bit3, FV_counter};
	end
	else if (i_rend_opt_read) begin
		o_vram_read = 1'b1;
		o_vram_addr = {1'b0, i_rend_opt_ptab, i_rend_opt_tilenum, i_rend_opt_bit3, i_rend_opt_fv};
	end
end
	

// Data delivery back to CPU
always @* begin
	o_cpu_datavalid = 1'b0;
	o_cpu_read_data = 8'bx;
	
	if (cpu_read_r) begin
		o_cpu_datavalid = 1'b1;
		
		case (cpu_addr_r)
			3'd2 : begin
				o_cpu_read_data[5] = spritelimit;
				o_cpu_read_data[6] = primaryhit;
				o_cpu_read_data[7] = vblank_in_progress;
			end
			
			3'd3 : begin
				o_cpu_read_data = cpu_spr_addr;
			end
			
			3'd4 : begin
				o_cpu_read_data = i_spr_rd_data;
			end
			
			3'd7 : begin
				// If accessing the palette vs. the rest of VRAM address space
				if (cpu_access_pal) begin
					o_cpu_read_data = {2'b0, i_pal_rd_data};
				end
				else begin
					o_cpu_read_data = vram_to_cpu_data_reg;
				end
			end
		endcase
	end
end





// Counter control
always @ (posedge i_clk or posedge i_reset) begin
	if (i_reset) begin
		FV_counter <= 'd0;
		V_counter <= 'd0;
		H_counter <= 'd0;
		VT_counter <= 'd0;
		HT_counter <= 'd0;	
	end
	else begin
		if (i_cpu_write && i_cpu_addr==3'd6 && cpu_secondwrite) begin
			FV_counter <= FV;
			V_counter <= V;
			H_counter <= H;
			VT_counter <= {VT[4:3], i_cpu_write_data[7:5]}; // VT [2:0] has old bits
			HT_counter <= i_cpu_write_data[4:0]; // HT contains old value
		end	
		else if (i_all_cnt_load) begin
			FV_counter <= FV;
			V_counter <= V;
			H_counter <= H;
			VT_counter <= VT;
			HT_counter <= HT;
		end
		else begin
			// Increment 15-bit counter by 1 or 32 on CPU access to 2007
			if (cpu_access_2007) begin
				{FV_counter, V_counter, H_counter, VT_counter, HT_counter} <= 
					{FV_counter, V_counter, H_counter, VT_counter, HT_counter} +
						((inc_port7_by_32)? 15'd32 : 15'd1);
			end
			else begin
				// Increment renderer HSCROLL counter
				if (i_hscroll_cnt_load) begin
					H_counter <= H;
					HT_counter <= HT;	
				end
				else if (i_hscroll_cnt_enable) begin
					{H_counter, HT_counter} <= {H_counter, HT_counter} + 6'd1;
				end
				
				// Increment renderer VSCROLL counter (VT register counts up to 29)
				if (i_vscroll_cnt_enable) begin
					if (VT_counter == 5'd29 && FV_counter == 3'b111) begin
						V_counter <= V_counter + 1'd1;
						VT_counter <= 5'd0;
						FV_counter <= 3'd0;
					end
					else begin
						{VT_counter, FV_counter} <= {VT_counter, FV_counter} + 8'd1;
					end
				end
			end
		end
	end
end

endmodule
